from mricluster.extractor import extract_worker_factory
from mricluster.reduce import reduce_worker_factory
from mricluster.cluster import cluster_worker_factory

import os
from multiprocessing import Process, Pool
import multiprocessing


class PipeLineMaser(object):
    """
    Note:
        Use this class with Tensorflow-CPU version as backend.
        Not compatible with Tensorflow-GPU version.
    """
    def __init__(self, query_dir, out_dir):
        """
        
        :param query_dir: The directory that contains the query images
        :param out_dir: The direcotry that outputs the result
        """
        self.train_set = self.populate_train_set(query_dir)
        self.out_dir = out_dir

    def populate_train_set(self, query_dir):
        """
        Search through the query directory and collects the image files into
        train data set.
        :param query_dir:
        :return:
        """
        res = []
        for root, dirs, files in os.walk(query_dir):
            for file in files:
                if file.endswith("jpg") or file.endswith("png"):
                    res.append(os.path.join(root, file))
        return res

    def pipe_line_worker(self, workload, verbose):

        extractWorker = extract_worker_factory(verbose=verbose)

        if verbose:
            print("[PipWorker] (pid=%d) Instantiate extract worker. workload: %d"
                  % (multiprocessing.current_process().pid, len(workload)))

        res = []
        for img in workload:
            features = extractWorker.extract_feature(img)
            res.append(features)

        return res


    def run(self, n_worker=-1, verbose=False):
        """

        :param n_worker:
        :return:
        """

        # Default setting: let number of workers equal to number of cpus
        if n_worker==-1:
            n_worker = multiprocessing.cpu_count()
        
        workload_count = len(self.train_set)

        """
            Extract Features
        """
        # Map
        process_pool = Pool(n_worker)
        extract_feature = process_pool.map(self.pipe_line_worker, self.train_set)

        if verbose:
            print ("Finish extracting features. Result matrix shape: (%d x %d)" % (len(extract_feature), len(extract_feature[0])))

        """
            Dimension Reduction
        """
        # TODO:



        """
            Clustering
        """
        # TODO:


def pipeline(input_dir, output_dir):
    """

    :param input_dir:
    :param output_dir:
    :return:
    """
    return PipeLineMaser(input_dir, output_dir)



from sklearn import (
    decomposition,
    manifold,
    pipeline,
)


class ReduceWorker(object):
    """
    Given a high-dimension feature arrays, reduce into low-dimension array
    """

    def __init__(self, model: manifold.TSNE):
        self.model = model

    def reduce(self, data: list[float]):
        """

        :param data: N x 2048 arrays
        :return:
        """
        res = self.model.fit_transform(data)
        return res


def __get_model(model: str, n_component: int):
    """
    PCA: Default n_components=48
    :param model:
    :param n_component: number of components to be reduced to
    :return:
    """
    if model == 'TSNE':
        return manifold.TSNE(random_state=0, verbose=1, n_components=n_component)
    if model == 'PCA-TSNE':
        tsne = manifold.TSNE(
            random_state=0, perplexity=50, early_exaggeration=6.0, n_components=n_component)
        pca = decomposition.PCA(n_components=48)
        return pipeline.Pipeline([('reduce_dims', pca), ('tsne', tsne)])
    if model == 'PCA':
        return decomposition.PCA(n_components=48)
    raise ValueError('Unknown model name')


def reduce_worker_factory(model: str="TSNE", n_component=2):
    """

    :param model: Name of the dimension reduce model. Default 'tSNE'
    :param n_component: number of components to be reduced to
    :return:
    """
    model = __get_model(model, n_component)
    return ReduceWorker(model)


import csv
import os

import numpy as np
import pandas
from keras import applications
from keras.applications.resnet50 import preprocess_input
from keras.preprocessing import image
from keras.engine.training import Model


class ModelWorker(object):
    """
    Given an image as a pixels array and a model, ouput the extracted features
    """

    def __init__(self, model: Model, verbose: bool=False):
        """
        :param img_path:
        :param model:
        :param verbose:
        """
        self.model = model
        self.verbose = verbose

    def extract_feature(self, img_path: str) -> list[float]:
        """

        :return: the extracted feature array of the given image.
        :rtype list[double]
        """
        # Load the image as the size of 224 x 224
        # TODO: Adjust img size by selected model input requirements
        img = image.load_img(img_path, target_size=(224, 224))


        # Convert image to numpy array
        x = image.img_to_array(img)
        x = np.expand_dims(x, axis=0)
        x = preprocess_input(x)

        # Extract the features
        features = self.model.predict(x)

        features = features[0]

        # Convert from np array to list[float]
        res = np.char.mod('%f', features)

        if self.verbose:
            print("[ExtractFeature] Predict feature shape: %d, %d" % (features.shape[0], features.shape[1]))
            print(len(res))

        return res


def __get_model(model: str):
    # include_top=False removes the fully connected layer at the end/top of the network
    # This allows us to get the feature vector as opposed to a classification
    if model == 'Xception':
        return applications.xception.Xception(weights='imagenet', include_top=False, pooling='avg')

    if model == 'VGG16':
        return applications.vgg16.VGG16(weights='imagenet', include_top=False, pooling='avg')

    if model == 'VGG19':
        return applications.vgg19.VGG19(weights='imagenet', include_top=False, pooling='avg')

    if model == 'InceptionV3':
        return applications.inception_v3.InceptionV3(weights='imagenet', include_top=False, pooling='avg')

    if model == 'MobileNet':
        return applications.mobilenet.MobileNet(weights='imagenet', include_top=False, pooling='avg')

    return applications.resnet50.ResNet50(weights='imagenet', include_top=False, pooling='avg')


def extract_worker_factory(model: str="ResNet50", verbose=False) -> ModelWorker:
    """
    :param model: the name of pre-trained model to use
    :return:
    """
    model = __get_model(model)
    return ModelWorker(model, verbose)


from matplotlib import pyplot as plt
import pandas
from sklearn.cluster import DBSCAN
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.datasets.samples_generator import make_blobs
import csv
import os
import shutil
import sys

class DBSCANWorker(object):
    """
    Cluster worker
    """

    def __init__(self, eps, min_samples):
        self.db = DBSCAN(eps=eps, min_samples=min_samples, n_jobs=-1)

    def fit(self, X) -> DBSCAN:
        """

        :param X: list[float]
        :return:
        """
        return self.db.fit(X)

    def labels(self):
        return self.db.labels_


def cluster_worker_factory(eps, min_samples):
    return DBSCAN(eps, min_samples)






from setuptools import setup

setup(name='mricluster',
      version='0.1',
      description='MRI image set clustering module based on Convolutional Neural Network model',
      long_description='',
      classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
      ],
      url='https://github.com/terrenceliu/MRI_Clustering',
      author='Terrence Liu',
      author_email='terrenceliu@rice.edu',
      license='MIT',
      packages=['mricluster'],
      install_requires= [
            'numpy',
            'pandas',
            'matplotlib',
            'tensorflow-gpu',
            'keras'
      ],
      zip_safe=False)